#include <fstream>
#include <iostream>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <cmath>
#include <bitset>
#include <time.h>
#include <fcntl.h>    /* For O_RDWR */

using namespace std;

inline

bool eval(const vector<vector<int> >& rules, 
          const bitset<50> & bs, int d) 
{
	bool coveredByRule;
	for(int i =0; i < rules.size(); i++) {
		coveredByRule = true;
		//cout << "rule size " << rules[i].size() << endl;
		for(int j=0; j < rules[i].size(); j++) {
			//cout << rules[i][j] << " " << bs[j] << endl;
			if(rules[i][j] != -1 && rules[i][j] !=  bs[j]) {
				coveredByRule = false;
				break;
			}// else {
			//	cout << "match" << endl;
			//}
		}

		if(coveredByRule) return 1;
					
	}
	return 0;
}

bool match(const vector<int>& rule1, const vector<int>& rule2) {

	int d =rule1.size();
	for (int j =0; j <d ; j++) {
		if(rule1[j] != rule2[j]) {
			return  false;						
		}	
	}
	return true;
}

bool matchOne(int indexRule, vector<vector<int> > rules) {

	for (int k=0; k < indexRule-1; k++){
		if(match(rules[indexRule],rules[k])) return true;
	}
	return false;

} 

int main(int argc, char** argv)
{
   unsigned k = atoi(argv[2]);
   int d = atoi(argv[1]);
   unsigned r = atoi(argv[3]); 
   float noise = atof(argv[4]);
   
   vector<vector<int> > rules;
   rules.resize(r);

   //int seed = (int) clock();
   int seed = time(NULL)*1000;

   int randomData = open("/dev/random",O_RDONLY);
   //int seed;
   read(randomData, &seed, sizeof seed);
   close(randomData);

   if (argc >= 6){
	seed = atoi(argv[5]);
   }
   srand(seed);

   for (int i=0; i < r; i++) 
   {
      	for(int j=0; j < k; j++) 
      	{
	 	rules[i].push_back(rand() % 2);	
      	}
      	for(int j=0; j < d-k; j++)
	{
		rules[i].push_back(-1);
	}

	random_shuffle(rules[i].begin(),rules[i].end());
	while(matchOne(i,rules)) {
		random_shuffle(rules[i].begin(),rules[i].end());
	
	}

   }

   cout << "% Seed " << seed << "\n";
   cout << "% Noise " << noise << "\n";

   bool first = true;
   for(int i=0; i < rules.size(); i++) {
	cout << "% Rule " << i << " ";
	first = true;
	for (int j =0; j < rules[i].size(); j++) {
		
		if(rules[i][j] != -1) {
			if (!first) cout << " && ";
			first = false;
			cout << "x" << j << "=" << rules[i][j];
		}	
	}
	cout << endl;
	
   }

   cout << "@relation kDNF problem \n\n";
   for(int i=0; i < d; i++) {
	cout << "@attribute x" << i << " {0,1}\n";
   }

   cout << "@attribute class {0,1}\n\n";
   cout << "@data\n";


   for( int i=0; i < pow((double)2,(int)d); i++) {
	bitset<50> bs((long) i);

	for (int j =0; j < d; j++) {
	 	cout << bs[j] << ",";	
	}
	//cout << bs.to_string();
	//cout << endl;

	bool e = eval(rules,bs,d);

	if (rand()/(float)RAND_MAX < noise) {
	  if(e) {
	    e = 0;
	  } else {
	    e = 1;
	  }
	}

	cout << e << endl;
	
   }
	

   return 0;
}
